/* jshint esversion: 6 */

const path = require('path');
const fs = require('fs');
const os = require('os');
const _ = require('lodash');
const yargs = require('yargs');
const yaml = require('yaml');
const boomack = require('boomack-js');

const packageInfo = require(path.join(__dirname, 'package.json'));

const mimeTypePattern = /^(?<type>[a-z]+)\/(?<subType>[\-a-z\d]+(?:\.[\-a-z\d]+)*(?:\+[a-z]+)?)(?:\s*;.+)?$/;

function mimeTypeCoersion(id) {
	const match = id.match(mimeTypePattern);
	if (match) {
		const type = match[1];
		const subType = match[2];
		return `${type}/${subType}`;
	}
	throw new Error("Invalid MIME type as media type ID.");
}

function generalSetup (y) {
	y.scriptName('boom');
	y.help();
	y.version(`Boomack CLI: ${packageInfo.version}${os.EOL}Boomack JS: ${boomack.version}`);
	y.options({
		'api': {
			describe: 'The base URL of the Boomack server',
			type: 'string',
		},
		'config': {
			describe: 'A path to a configuration file',
			type: 'array',
			normalize: true,
		},
		'token': {
			describe: 'An authentication token.',
			type: 'string',
		},
		'dry': {
			describe: 'Do not send, but print the request',
			type: 'boolean',
		},
		'curl': {
			describe: 'Do not send, but print the request in cURL syntax',
			type: 'boolean',
		},
		'silent': {
			describe: 'Do not show info messages',
			type: 'boolean',
		},
	});
	return y;
}

function defaultCommand (y) {
	y.command(['* [content]', 'display [content]'], 'display a media item', y => {
		y.positional('content', {
			describe: 'a path to a local file, a URL or an arbitrary string',
			type: 'string',
		});
		y.options({
			'l': {
				alias: ['location', 'target'],
				describe: 'target panel and slot ID, e.g. panel-1/slot-a',
				type: 'string',
			},
			's': {
				alias: 'string',
				describe: 'an arbitrary string to display',
				type: 'string',
				conflicts: ['f', 'u'],
			},
			'f': {
				alias: 'file',
				describe: 'a path to a file, with content to display',
				type: 'string',
				conflicts: ['s', 'u'],
			},
			'u': {
				alias: 'url',
				describe: 'a http, https, or data url of a resource to display',
				type: 'string',
				conflicts: ['s', 'f'],
			},
			'w': {
				alias: 'watch',
				describe: 'watch a file and display it, every time it changed',
				type: 'boolean',
				implies: 'f',
			},
			't': {
				alias: 'type',
				describe: 'the MIME type of the content',
				type: 'string',
			},
			'c': {
				alias: 'cache',
				describe: 'a switch to control the caching on the server',
				choices: ['auto', 'memory', 'file'],
			},
			'e': {
				alias: 'embed',
				describe: 'a switch to state the intention, that the media content should be embedded into the HTML and not referenced as resource',
				type: 'boolean',
			},
			'x': {
				alias: 'extend',
				describe: 'the position to add to existing content in the target slot',
				choices: ['start', 'end'],
			},
			'append': {
				describe: 'same as --extend end',
				type: 'boolean',
			},
			'prepend': {
				describe: 'same as --extend start',
				type: 'boolean',
			},
			'title': {
				describe: 'the title of the media item',
				type: 'string',
			},
			'o': {
				alias: 'options',
				describe: 'one or more options in key=value form, as YAML/JSON encoded object, or STDIN',
				type: 'array',
			},
			'p': {
				alias: 'presets',
				describe: 'one or more IDs of presets',
				type: 'array',
			},
			'r': {
				alias: 'raw',
				describe: 'send content directly as request. ignores all options but --string, --file, and --url',
				type: 'boolean',
			},
			'no-file-url': {
				describe: 'prevent passing a file:// URL if file is displayed on localhost',
				type: 'boolean',
			},
			'debug': {
				describe: 'show debug information instead of actual media item',
				type: 'boolean',
			},
		});
	});
	return y;
}

function evalCommand (y) {
	y.command('eval [code]', 'evaluate JavaScript code in a panel', y => {
		y.positional('code', {
			describe: 'a path to a local file, a URL, or JavaScript code to evaluate',
			type: 'string',
		});
		y.options({
			'l': {
				alias: ['location', 'target'],
				describe: 'target panel ID',
				type: 'string',
			},
		});
	});
	return y;
}

const exportOptions = {
	'p': {
		alias: 'path',
		describe: 'a relative path to export to',
		type: 'string',
	},
	'n': {
		alias: 'name',
		describe: 'a name for the exported HTML file; without extension',
		type: 'string',
	},
	't': {
		alias: 'theme',
		describe: 'the theme to use during the export',
		type: 'string',
	},
	'z': {
		alias: 'zoom',
		describe: 'a CSS zoom level for the HTML',
		type: 'number',
	},
};

const exportPanelOptions = {
	...exportOptions,
	'no-logo': {
		describe: 'removes the logo from the panel header during export',
		type: 'boolean',
	},
	'no-header': {
		describe: 'removes the panel header during export',
		type: 'boolean',
	},
	'no-toolbars': {
		describe: 'removes the toolbars from all slots during export',
		type: 'boolean',
	},
	'no-tools': {
		describe: 'removes all tools from the slot toolbars during export',
		type: 'boolean',
	},
};

const exportSlotOptions = {
	...exportOptions,
	'no-toolbar': {
		describe: 'removes the toolbar from the slot during export',
		type: 'boolean',
	},
	'no-tools': {
		describe: 'removes all tools from the slot\'s toolbar during export',
		type: 'boolean',
	},
};

function exportCommand (y) {

	function allSubCommand (y) {
		y.command(['*', 'all'],
			'export all panels to the filesystem on the server',
			y => {
				y.options(_.omit(exportPanelOptions, 'n'));
			});
		return y;
	}

	function panelSubCommand (y) {
		y.command(['panel [id]'],
			'export one panel to the filesystem on the server',
			y => {
				y.positional('id', {
					describe: 'the ID of the panel',
					type: 'string',
					default: 'default',
				});
				y.options(exportPanelOptions);
			});
		return y;
	}

	function slotSubCommand (y) {
		y.command(['slot <target>'],
			'export a slot to the filesystem on the server',
			y => {
				y.positional('target', {
					describe: 'the panel and slot ID, e.g. panel-1/main-slot',
					type: 'string',
				});
				y.options(exportSlotOptions);
			});
		return y;
	}

	const subCommands = [
		allSubCommand,
		panelSubCommand,
		slotSubCommand,
	];
	y.command('export', 'export to server filesystem',
		y2 => _.reduce(subCommands, (y3, cmd) => cmd(y3), y2));
	return y;
}

function panelCommand (y) {

	function lsSubCommand (y) {
		 y.command(['*', 'ls'],
			'list existing panels',
			y => {
			});
		return y;
	}

	function getSubCommand (y) {
		y.command('get [id]',
			'get the structure of a panel',
			y => {
				y.positional('id', {
					describe: 'the ID of the panel',
					type: 'string',
					default: 'default',
				});
			});
		return y;
	}

	function addSubCommand (y) {
		 y.command('add <id> [layout]',
			'create a new panel',
			y => {
				y.positional('id', {
					describe: 'the ID of the panel to create',
					type: 'string',
				});
				y.positional('layout', {
					describe: 'a path to a JSON/YAML file, a JSON encoded layout request, or STDIN',
					type: 'string',
				});
			});
		return y;
	}

	function layoutSubCommand (y) {
		 y.command(['layout <id> <layout>', 'update <id> <layout>'],
			'change the layout of a panel',
			y => {
				y.positional('id', {
					describe: 'the ID of the panel to update',
					type: 'string',
				});
				y.positional('layout', {
					describe: 'a path to a JSON/YAML file, a JSON encoded layout request, or STDIN',
					type: 'string',
				});
			});
		return y;
	}

	function clearSubCommand (y) {
		y.command('clear [id]',
			'clear a all slots in a panel',
			y => {
				y.positional('id', {
					describe: 'the ID of the panel',
					type: 'string',
					default: 'default',
				});
			});
		return y;
	}

	function exportSubCommand (y) {
		y.command('export [id]',
			'export a panel to the filesystem on the server',
			y => {
				y.positional('id', {
					describe: 'the ID of the panel',
					type: 'string',
					default: 'default',
				});
				y.options(exportPanelOptions);
			});
		return y;
	}

	function deleteSubCommand (y) {
		 y.command('delete <id>',
			'delete a panel',
			y => {
				y.positional('id', {
					describe: 'the ID of the panel to delete',
					type: 'string',
				});
			});
		return y;
	}

	const subCommands = [
		lsSubCommand,
		getSubCommand,
		addSubCommand,
		layoutSubCommand,
		clearSubCommand,
		exportSubCommand,
		deleteSubCommand,
	];
	y.command('panel', 'manage panels',
		y2 => _.reduce(subCommands, (y3, cmd) => cmd(y3), y2));
	return y;
}

function slotCommand (y) {

	function clearSubCommand (y) {
		y.command('clear <target>',
		   'clear a slot',
		   y => {
			   y.positional('target', {
				   describe: 'the panel and slot ID, e. g. panel-1/main-slot',
				   type: 'string',
			   });
		   });
	   return y;
   }

   function exportSubCommand (y) {
		y.command('export <target>',
			'export a slot to the server filesystem',
			y => {
				y.positional('target', {
					describe: 'the panel and slot ID, e. g. panel-1/main-slot',
					type: 'string',
				});
				y.options(exportSlotOptions);
			});
		return y;
	}

   const subCommands = [
	   clearSubCommand,
	   exportSubCommand,
   ];
   y.command('slot', 'manage slots',
	   y2 => _.reduce(subCommands, (y3, cmd) => cmd(y3), y2));
   return y;
}

function presetCommand (y) {

	function lsSubCommand (y) {
		 y.command(['*', 'ls'],
			'list existing presets',
			y => {
			});
		return y;
	}

	function getSubCommand(y) {
		y.command('get <id>',
			'get an existing preset',
			y => {
				y.positional('id', {
					describe: 'the ID of the preset',
					type: 'string',
				});
			});
		return y;
	}

	function addSubCommand (y) {
		 y.command('add <id> [options..]',
			'create a new preset',
			y => {
				y.positional('id', {
					describe: 'the ID of the preset to create',
					type: 'string',
				});
				y.positional('options', {
					describe: 'path(s) to JSON/YAML file, JSON encoded presets, one or more options in key=value form, or STDIN',
					type: 'string',
				});
			});
		return y;
	}

	function updateSubCommand (y) {
		 y.command('update <id> [options..]',
			'update an preset',
			y => {
				y.positional('id', {
					describe: 'the ID of the preset to update',
					type: 'string',
				});
				y.positional('options', {
					describe: 'path(s) to JSON/YAML file, JSON encoded presets, one or more options in key=value form, or STDIN',
					type: 'string',
				});
			});
		return y;
	}

	function deleteSubCommand (y) {
		 y.command('delete <id>',
			'delete an preset',
			y => {
				y.positional('id', {
					describe: 'the ID of the preset to delete',
					type: 'string',
				});
			});
		return y;
	}

	const subCommands = [
		lsSubCommand,
		getSubCommand,
		addSubCommand,
		updateSubCommand,
		deleteSubCommand,
	];
	y.command('preset', 'manage presets',
		y2 => _.reduce(subCommands, (y3, cmd) => cmd(y3), y2));
	return y;
}

function mediaTypeCommand (y) {

	function lsSubCommand (y) {
		 y.command(['*', 'ls'],
			'list existing media types',
			y => {
			});
		return y;
	}

	function getSubCommand(y) {
		y.command('get <id>',
			'get an existing media type',
			y => {
				y.positional('id', {
					describe: 'the ID of the media type',
					type: 'string',
					coerce: mimeTypeCoersion,
				});
			});
		return y;
	}

	function addSubCommand (y) {
		 y.command('add <id> [options..]',
			'create a new media type',
			y => {
				y.positional('id', {
					describe: 'the ID of the media type to create',
					type: 'string',
					coerce: mimeTypeCoersion,
				});
				y.positional('options', {
					describe: 'path(s) to JSON/YAML file, JSON encoded presets, one or more options in key=value form, or STDIN',
					type: 'string',
					default: '',
				});
				y.options({
					'p': {
						alias: 'presets',
						describe: 'one or more IDs of presets',
						type: 'array',
					},
					'text': {
						describe: 'indicate this type to be a character sequence',
						type: 'boolean',
					},
				});
			});
		return y;
	}

	function updateSubCommand (y) {
		 y.command('update <id> [options..]',
			'update a media type',
			y => {
				y.positional('id', {
					describe: 'the ID of the preset to update',
					type: 'string',
					coerce: mimeTypeCoersion,
				});
				y.positional('options', {
					describe: 'path(s) to JSON/YAML file, JSON encoded presets, one or more options in key=value form, or STDIN',
					type: 'string',
					default: '',
				});
				y.options({
					'p': {
						alias: 'presets',
						describe: 'one or more IDs of presets',
						type: 'array',
					},
					'text': {
						describe: 'indicate this type to be a character sequence',
						type: 'boolean',
					},
				});
			});
		return y;
	}

	function deleteSubCommand (y) {
		 y.command('delete <id>',
			'delete a media type',
			y => {
				y.positional('id', {
					describe: 'the ID of the media type to delete',
					type: 'string',
					coerce: mimeTypeCoersion,
				});
			});
		return y;
	}

	const subCommands = [
		lsSubCommand,
		getSubCommand,
		addSubCommand,
		updateSubCommand,
		deleteSubCommand,
	];
	y.command('type', 'manage media types',
		y2 => _.reduce(subCommands, (y3, cmd) => cmd(y3), y2));
	return y;
}

function actionCommand (y) {

	function lsSubCommand (y) {
		 y.command(['*', 'ls'],
			'list existing actions',
			y => {
			});
		return y;
	}

	function getSubCommand(y) {
		y.command('get <id>',
			'get an existing action',
			y => {
				y.positional('id', {
					describe: 'the ID of the action',
					type: 'string',
				});
			});
		return y;
	}

	function addSubCommand (y) {
		 y.command('add <id> <action>',
			'create a new action',
			y => {
				y.positional('id', {
					describe: 'the ID of the action to create',
					type: 'string',
				});
				y.positional('action', {
					describe: 'a path to JSON/YAML file, a JSON encoded action, or STDIN',
					type: 'string',
				});
			});
		return y;
	}

	function updateSubCommand (y) {
		 y.command('update <id> <action>',
			'update an action',
			y => {
				y.positional('id', {
					describe: 'the ID of the action to update',
					type: 'string',
				});
				y.positional('action', {
					describe: 'a path to JSON/YAML file, a JSON encoded action, or STDIN',
					type: 'string',
				});
			});
		return y;
	}

	function deleteSubCommand (y) {
		 y.command('delete <id>',
			'delete an action',
			y => {
				y.positional('id', {
					describe: 'the ID of the action to delete',
					type: 'string',
				});
			});
		return y;
	}

	const subCommands = [
		lsSubCommand,
		getSubCommand,
		addSubCommand,
		updateSubCommand,
		deleteSubCommand,
	];
	y.command('action', 'manage actions',
		y2 => _.reduce(subCommands, (y3, cmd) => cmd(y3), y2));
	return y;
}

function detectCommand(argv) {
	const p1 = _.get(argv._, 0);
	const p2 = _.get(argv._, 1);
	switch (p1) {
		case 'eval':
			return 'eval';
		case 'panel':
			if (_.isNil(p2)) return `${p1} ls`;
			if (_.includes(['ls', 'get', 'add', 'update', 'layout', 'clear', 'export', 'delete'], p2)) return `${p1} ${p2}`;
			throw new Error("Invalid sub-command for command panel: " + p2);
		case 'slot':
			if (_.isNil(p2)) throw new Error("Sub-command needed for command slot.");
			if (_.includes(['clear', 'export'], p2)) return `${p1} ${p2}`;
			throw new Error("Invalid sub-command for command slot: " + p2);
		case 'preset':
			if (_.isNil(p2)) return `${p1} ls`;
			if (_.includes(['ls', 'get', 'add', 'update', 'delete'], p2)) return `${p1} ${p2}`;
			throw new Error("Invalid sub-command for command options: " + p2);
		case 'type':
			if (_.isNil(p2)) return `${p1} ls`;
			if (_.includes(['ls', 'get', 'add', 'update', 'delete'], p2)) return `${p1} ${p2}`;
			throw new Error("Invalid sub-command for command options: " + p2);
		case 'action':
			if (_.isNil(p2)) return `${p1} ls`;
			if (_.includes(['ls', 'get', 'add', 'update', 'delete'], p2)) return `${p1} ${p2}`;
			throw new Error("Invalid sub-command for command action: " + p2);
		case 'export':
			if (_.isNil(p2)) return `${p1} all`;
			if (_.includes(['all', 'panel', 'slot'], p2)) return `${p1} ${p2}`;
			throw new Error("Invalid sub-command for command export: " + p2);
		default:
			return 'display';
	}
}

function detectStdin(argv, command) {
	if (command !== 'display' && command !== 'eval') return undefined;
	return _.isNil(argv.string) &&
		_.isNil(argv.code) &&
		(_.isNil(argv.file) || argv.file === '-') &&
		_.isNil(argv.url) &&
		_.isNil(argv.content);
}

const dataUrlPattern = /^data:([a-z]+\/[a-zA-Z0-9+-]+(\s*;\s*charset=.+?)?(\s*;\s*base64)?)?\s*,.+/;

function detectUrlInput(argv, command, stdin) {
	if (command !== 'display' && command !== 'eval') return undefined;
	if (stdin) return null;
	if (argv.string) return null;
	if (argv.file) return null;
	if (argv.url) return argv.url;
	const arg = argv.content || argv.code;
	if (!_.isString(arg)) return null;
	if (_.startsWith(arg, 'http://') ||
		_.startsWith(arg, 'https://') ||
		arg.match(dataUrlPattern)) return arg;
	return null;
}

function detectFilePathInput(argv, command, stdin, url) {
	if (command !== 'display' && command !== 'eval') return undefined;
	if (stdin) return null;
	if (url) return null;
	if (argv.string) return null;
	if (argv.file) return argv.file;
	const arg = argv.content || argv.code;
	if (!_.isString(arg)) return null;
	try {
		fs.accessSync(arg, fs.constants.F_OK);
		return arg;
	} catch (e) {
		return null;
	}
}

function detectStringInput(argv, command, stdin, url, file) {
	if (command !== 'display') return undefined;
	if (stdin) return null;
	if (url) return null;
	if (file) return null;
	if (argv.string) return argv.string;
	const arg = argv.content;
	if (!_.isString(arg)) return null;
	return arg;
}

function parseTarget(argv, command) {
	if (command === 'eval') {
		return [argv.target, undefined];
	}
	if (command !== 'display' && command !== 'slot clear' && command !== 'slot export' && command !== 'export slot') {
		return [undefined, undefined];
	}
	const target = argv.target;
	if (_.isString(target)) {
		if (_.includes(target, '/')) {
			let [ panelId, slotId ] = target.split(/\//);
			return [panelId || null, slotId || null];
		} else {
			return [null, target];
		}
	} else {
		return [null, null];
	}
}

function parseCache(argv) {
	if (argv.cache === 'auto') return null;
	return argv.cache || null;
}

function parseExtend(argv) {
	if (argv.extend) return argv.extend;
	if (argv.append) return 'end';
	if (argv.prepend) return 'start';
	return null;
}

let stdinConsumed = false;

function parseStructure(v, term, parser) {
	if (_.isNil(v)) return null;
	if (_.isString(v) && _.toLower(v) === 'stdin' && !stdinConsumed) {
		try {
			v = fs.readFileSync(process.stdin.fd, 'utf-8');
		} catch (e) {
			console.log("Reading from STDIN failed.");
			process.exit(1);
			return null;
		}
		stdinConsumed = true;
	} else {
		let fileOK = null;
		try {
			fs.accessSync(v, fs.constants.R_OK);
			fileOK = true;
		} catch (e) { }
		if (fileOK) {
			v = fs.readFileSync(v, 'utf-8');
		}
	}
	let yamlErr = null;
	try {
		v = yaml.parse(v);
	} catch (e) {
		yamlErr = e;
	}
	if (_.isString(v) || yamlErr) {
		if (parser) {
			try {
				return parser(v);
			} catch (e) {
				console.log('No valid expression for ' + term + ':', parserErr.message);
				return null;
			}
		} else if (yamlErr) {
			console.log('No valid JSON/YAML for ' + term + ':', yamlErr.message);
			return null;
		} else {
			return null;
		}
	} else {
		return v;
	}
}

function parseLayout(argv) {
	// TODO implement layout mini-language
	const layoutExpressionParser = null;
	return parseStructure(argv.layout, 'layout', layoutExpressionParser);
}

function parseOptions(argv) {
	let options = [];

	function keyValueParser(v) {
		const result = {};
		v = v.split(/\n|\r|\r\n/g);
		_.forEach(v, l => {
			l = _.trim(l);
			const s = l.indexOf('=')
			if (s > 0) {
				result[l.substr(0, s)] = l.substr(s + 1);
			}
		});
		return result;
	}

	_.forEach(argv.options, v => {
		const o = parseStructure(v, 'options', keyValueParser);
		if (o) options.push(o);
	});
	return _.reduce(options, (o1, o2) => _.assign(o1, o2), {});
}

function parseAction(argv) {
	// TODO implement action mini-language
	const actionExpressionParser = null;
	return parseStructure(argv.action, 'action', actionExpressionParser);
}

exports.parse = function (args) {
	const commands = [
		generalSetup,
		defaultCommand,
		evalCommand,
		panelCommand,
		slotCommand,
		presetCommand,
		mediaTypeCommand,
		actionCommand,
		exportCommand,
	];
	const parser = _.reduce(commands, (y, cmd) => cmd(y), yargs);
	const argv = parser.parse(args);
	const command = detectCommand(argv);
	const [ panelId, slotId ] = parseTarget(argv, command);
	const stdin = detectStdin(argv, command);
	const url = detectUrlInput(argv, command, stdin);
	const file = detectFilePathInput(argv, command, stdin, url);
	const string = detectStringInput(argv, command, stdin, url, file);
	const task = {
		// common
		command: command,
		api: argv.api,
		configFiles: argv.config || [],
		token: argv.token,
		silent: argv.silent || argv.curl || false,
		dry: argv.dry || argv.curl || false,
		curl: argv.curl || false,
		// display
		stdin: stdin,
		string: string,
		file: file,
		url: url,
		noFileUrl: argv.fileUrl === false,
		watch: command === 'display' ? argv.watch : undefined,
		type: command === 'display' ? argv.type : undefined,
		cache: command === 'display' ? parseCache(argv) : undefined,
		embed: command === 'display' ? argv.embed : undefined,
		extend: command === 'display' ? parseExtend(argv, command) : undefined,
		title: command === 'display' ? argv.title : undefined,
		debug: command === 'display' ? argv.debug : undefined,
		presets: argv.presets,
		// display, panel clear
		panelId: panelId,
		slotId: slotId,
		// display, preset add, preset update
		options: parseOptions(argv),
		// type add, type update
		text: argv.text || false,
		// eval
		code: argv.code,
		// panel *, preset *, type *, action *
		id: argv.id,
		// export *
		path: argv.path,
		theme: argv.theme,
		zoom: argv.zoom,
		// export panel, export slot
		name: argv.name,
		logo: argv.logo !== false,
		header: argv.header !== false,
		toolbars: argv.toolbars !== false,
		toolbar: argv.toolbar !== false,
		tools: argv.tools !== false,
		// panel add, panel layout
		layout: parseLayout(argv),
		// action
		action: parseAction(argv),
	};
	return task;
};
