/* jshint esversion: 8 */

const { URL, pathToFileURL } = require('url');
const path = require('path');
const fs = require('fs');
const _ = require('lodash');
const { Address4, Address6 } = require('ip-address');
const boomack = require('boomack-js');
const types = require('./types');

function isStream(x) {
	return _.isFunction(x.pipe) && _.isFunction(x.on);
}

function isServerLocal(b) {
	const url = new URL(b.apiUrl);
	if (url.hostname === 'localhost') return true;
	try {
		const ip4 = new Address4(url.hostname);
		if (ip4.correctForm() === '127.0.0.1') {
			return true;
		}
	} catch (err) {
		// ignore
	}
	try {
		const ip6 = Address6.fromURL(b.apiUrl);
		if ((ip6.is4() && ip6.to4().correct === '127.0.0.1') || (ip6.correct === '::1')) {
			return true;
		}
	} catch (err) {
		// ignore
	}
	return false;
}

function guessTypeFromBinary(data) {
	if (data.slice(0, 8).compare(Buffer.from([0x89, 0x50, 0x4e, 0x47, 0x0d, 0x0a, 0x1a, 0x0a])) == 0) {
		return 'image/png';
	}
	const head6Str = data.slice(0, 6).toString('ascii');
	if (head6Str == 'GIF87a' || head6Str == 'GIF89a') {
		return 'image/gif';
	}
	if (data.slice(0, 2).compare(Buffer.from([0xff, 0xd8])) == 0 &&
		data.slice(data.length - 2, data.length).compare(Buffer.from([0xff, 0xd9])) == 0) {
		return 'image/jpeg';
	}
	var sample = data.slice(0, 1024);
	var sampleText = sample.toString('ascii');
	return guessTypeFromString(sampleText);
}

function guessTypeFromString(txt) {
	txt = _.trim(txt);
	const TXT = txt.toUpperCase();
	if (_.startsWith(TXT, '<!DOCTYPE HTML>') ||
		_.startsWith(TXT, '<HTML') && _.endsWith(txt, '</HTML>')) {
		return 'text/html';
	}
	if (_.startsWith(txt, '<?xml ') && TXT.includes('<SVG') && TXT.includes('</SVG>') ||
		_.startsWith(TXT, '<SVG') && _.endsWith('</SVG>')) {
		return 'image/svg+xml';
	}
	if (_.startsWith(txt, '<') && _.endsWith(txt, '>')) {
		return 'text/html';
	}
	return 'text/plain';
}

async function display(b, argv) {
	const request = { };
	let options = { };

	// set target
	if (argv.panelId) request.panel = argv.panelId;
	if (argv.slotId) request.slot = argv.slotId;

	let binaryData = null;

	// set title
	if (argv.title) {
		request.title = argv.title;
	} else if (argv.file) {
		// from file path
		request.title = path.basename(argv.file);
	} else if (argv.url) {
		let url;
		try {
			url = new URL(argv.url);
		} catch (err) {
			console.error("Invalid URL:", argv.url);
			process.exit(1);
			return;
		}
		if (url.protocol !== 'http' &&
			url.protocol !== 'https' &&
			url.protocol !== 'file'
		) {
			console.error("Invalid protocol in URL. Allowed: http, https, file.");
			process.exit(1);
			return;
		}
		request.title =  url.host + url.pathname;
		if (url.search) {
			request.title += '?...';
		}
	}

	// get or guess type
	if (argv.type) {
		request.type = argv.type;
	} else {
		if (argv.file) {
			request.type = types.typeFromPath(b.config.client.types, argv.file);
			if (request.type && !argv.silent) {
				console.log("Guessed the type from filename: " + request.type);
				console.log("Hope that is OK. If not, specify type with -t <MIME type>.");
			}
		} else if (argv.url) {
			const url = new URL(argv.url);
			request.type = types.typeFromPath(b.config.client.types, url.pathname);
			if (request.type && !argv.silent) {
				console.log("Guessed the type from URL path: " + request.type);
				console.log("Hope that is OK. If not, specify type with -t <MIME type>.");
			} else {
				request.type = await boomack.getMediaTypeOfUrl(argv.url);
				if (request.type && !argv.silent) {
					console.log("Got the type with a HEAD request on the URL: " + request.type);
					console.log("Hope that is OK. If not, specify type with -t <MIME type>.");
				}
			}
		} else if (argv.string) {
			request.type = guessTypeFromString(argv.string);
			if (request.type && !argv.silent) {
				console.log("Guessed the type from text content: " + request.type);
				console.log("Hope that is OK. If not, specify type with -t <MIME type>.");
			}
		} else if (binaryData) {
			request.type = guessTypeFromBinary(binaryData);
			if (request.type && !argv.silent) {
				console.log("Guessed the type from binary content: " + request.type);
				console.log("Hope that is OK. If not, specify type with -t <MIME type>.");
			}
		}
	}
	if (!request.type) {
		if (!argv.silent) console.log('No type given. Canceling the request.');
		process.exit(1);
		return;
	}

	// options
	_.assign(options, argv.options);
	if (argv.cache) options.cache = argv.cache;
	if (argv.embed) options.embed = true;
	if (argv.extend) options.extend = argv.extend;
	if (argv.debug) options.debug = true;

	// load content
	let useFileStream = false;
	if (argv.stdin) {
		if (argv.watch) {
			console.log('You can not watch STDIN. Watch flag is ignored.');
		}
		try {
			request.stream = process.stdin;
		} catch (e) {
			if (!argv.silent) console.log("Reading from STDIN failed.");
			console.error(e);
			process.exit(1);
			return;
		}
	} else if (argv.string) {
		request.text = argv.string;
	} else if (argv.file) {
		if (!argv.noFileUrl && isServerLocal(b)) {
			request.src = pathToFileURL(path.resolve(argv.file));
		} else {
			useFileStream = true;
		}
	} else if (argv.url) {
		request.src = argv.url;
	}

	async function postContent() {
		if (useFileStream) {
			const stat = await fs.promises.stat(argv.file);
			request.streamLength = stat.size;
			request.stream = fs.createReadStream(argv.file);
		}
		if (request.stream) {
			if (!request.panel) request.panel = 'default';
			let response = null;
			if (request.slot) {
				response = await b.streamMediaItemToSlot(request.panel, request.slot,
					request.type, request.stream, request.streamLength,
					request.title, argv.presets, options);
			} else {
				response = await b.streamMediaItemToPanel(request.panel,
					request.type, request.stream, request.streamLength,
					request.title, argv.presets, options);
			}
			if (!argv.stdin) {
				request.stream.destroy();
			}
			return response;
		} else {
			if (!_.isEmpty(argv.presets)) {
				request.presets = argv.presets;
			}
			if (!_.isEmpty(options)) {
				request.options = options;
			}
			return await b.displayMediaItems(request);
		}
	}

	if (argv.watch && !argv.stdin && !argv.dry) {
		// watch file if watch option is set

		let busy = false;
		let changedWhileBusy = false;
		let resolveCancelPromise = null;
		const cancelPromise = new Promise((resolve, reject) => {
			resolveCancelPromise = resolve;
		});

		async function redisplay() {
			try {
				const response = await postContent();
				if (!response.success) {
					console.error('Displaying file content failed with HTTP status', response.status);
				}
			} catch (err) {
				console.error('Displaying file content failed with error:', err);
			}
		}

		function onFileChanged() {
			if (busy) {
				changedWhileBusy = true;
				return;
			}
			busy = true;
			redisplay()
				.then(() => {
					busy = false;
					if (changedWhileBusy) {
						changedWhileBusy = false;
						onFileChanged();
					}
				})
				.catch(err => {
					console.error('Failed to open source file for reading:', err);
				});
		}

		const chokidar = require('chokidar');
		const watcher = chokidar.watch(argv.file, {
			persistent: true,
			ignoreInitial: false,
			followSymlinks: true,
			disableGlobbing: true,
			atomic: 100,
		});
		watcher.on('add', onFileChanged);
		watcher.on('change', onFileChanged);
		watcher.on('ready', () => {
			console.log('Watching file', argv.file, '...');
			console.log('Press Ctrl+C to stop');
		});

		function cancel() {
			watcher.close().then(() => {
				console.log('Stopped watching');
				resolveCancelPromise();
			});
		}

		process.on('SIGINT', cancel);
		process.on('SIGTERM', cancel);

		await cancelPromise;
		return null;
	} else {
		// otherwise; just post the file content
		return await postContent();
	}
}

async function evaluateCode(b, argv) {
	const request = { };

	// set target
	if (argv.panelId) request.panel = argv.panelId;

	// load content
	if (argv.stdin) {
		try {
			request.stream = process.stdin;
		} catch (e) {
			if (!argv.silent) console.log("Reading from STDIN failed.");
			console.error(e);
			process.exit(1);
			return;
		}
	} else if (argv.file) {
		const stat = await fs.promises.stat(argv.file);
		request.streamLength = stat.size;
		request.stream = fs.createReadStream(argv.file);
	} else if (argv.code) {
		request.script = argv.code;
	}

	if (request.stream) {
		if (!request.panel) request.panel = 'default';
		return await b.streamCodeForEvaluation(request.panel,
			request.stream, request.streamLength);
	} else {
		return await b.evaluateCode(request);
	}
}

function panelList(b, argv) {
	return b.listPanels();
}

function panelGet(b, argv) {
	const panelId = argv.id || 'default';
	return b.getPanel(panelId);
}

function panelAdd(b, argv) {
	const panelId = argv.id || 'default';
	const layout = argv.layout || {
		grid: { },
		slots: { default: { } },
	};
	return b.addPanel(panelId, layout);
}

function panelLayout(b, argv) {
	const panelId = argv.id || 'default';
	if (!argv.layout) {
		if (!argv.silent) console.log("No panel layout. Cancelling request.");
		process.exit(1);
		return;
	}
	return b.switchLayout(panelId, argv.layout);
}

function panelClear(b, argv) {
	const panelId = argv.id;
	return b.clearPanel(panelId);
}

function panelExport(b, argv) {
	const panelId = argv.id;
	const theme = argv.theme;
	const zoom = argv.zoom;
	const subPath = argv.path;
	const filename = argv.name;
	const logo = argv.logo;
	const header = argv.header;
	const tools = argv.tools;
	const toolbars = argv.toolbars;
	return b.exportPanel(panelId, {
		path: subPath,
		name: filename,
		theme, zoom, logo, header, tools, toolbars,
	});
}

function panelDelete(b, argv) {
	const panelId = argv.id;
	if (!panelId || panelId === 'default') {
		if (!argv.silent) console.log("Invalid panel ID. Cancelling request.");
		process.exit(1);
		return;
	}
	return b.deletePanel(panelId);
}

function slotClear(b, argv) {
	const panelId = argv.panelId || 'default';
	const slotId = argv.slotId;
	if (!slotId) {
		if (!argv.silent) console.log("No slot ID given. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.clearSlot(panelId, slotId);
}

function slotExport(b, argv) {
	const panelId = argv.panelId || 'default';
	const slotId = argv.slotId;
	if (!slotId) {
		if (!argv.silent) console.log("No slot ID given. Cancelling the request.");
		process.exit(1);
		return;
	}
	const theme = argv.theme;
	const zoom = argv.zoom;
	const subPath = argv.path;
	const filename = argv.name;
	const tools = argv.tools;
	const toolbar = argv.toolbar;
	return b.exportSlot(panelId, slotId, {
		path: subPath,
		name: filename,
		theme, zoom, tools, toolbar,
	});
}

function presetList(b, argv) {
	return b.listPresets();
}

function presetGet(b, argv) {
	const presetId = argv.id;
	if (!presetId) {
		if (!argv.silent) console.log("No ID given for preset. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.getPreset(argv.id);
}

function presetAdd(b, argv) {
	const presetId = argv.id;
	if (!presetId) {
		if (!argv.silent) console.log("No ID given for preset. Cancelling the request.");
		process.exit(1);
		return;
	}
	const options = argv.options;
	if (_.isEmpty(options)) {
		if (!argv.silent) console.log("Options are empty. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.addPreset(presetId, options);
}

function presetUpdate(b, argv) {
	const presetId = argv.id;
	if (!presetId) {
		if (!argv.silent) console.log("No ID given for preset. Cancelling the request.");
		process.exit(1);
		return;
	}
	const options = argv.options;
	if (_.isEmpty(options)) {
		if (!argv.silent) console.log("Options are empty. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.updatePreset(presetId, options);
}

function presetDelete(b, argv) {
	const presetId = argv.id;
	if (!presetId) {
		if (!argv.silent) console.log("No ID given for preset. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.deletePreset(presetId);
}

function mediaTypeList(b, argv) {
	return b.listMediaTypes();
}

function mediaTypeGet(b, argv) {
	const mediaTypeId = argv.id;
	if (!mediaTypeId) {
		if (!argv.silent) console.log("No ID given for media type. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.getMediaType(argv.id);
}

function mediaTypeAdd(b, argv) {
	const mediaTypeId = argv.id;
	if (!mediaTypeId) {
		if (!argv.silent) console.log("No ID given for media type. Cancelling the request.");
		process.exit(1);
		return;
	}
	const text = argv.text;
	const presets = argv.presets || [];
	const options = argv.options;
	if (_.isEmpty(options)) {
		if (!argv.silent) console.log("Options are empty. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.addMediaType(mediaTypeId, { text, presets, options });
}

function mediaTypeUpdate(b, argv) {
	const mediaTypeId = argv.id;
	if (!mediaTypeId) {
		if (!argv.silent) console.log("No ID given for media type. Cancelling the request.");
		process.exit(1);
		return;
	}
	const text = argv.text;
	const presets = argv.presets || [];
	const options = argv.options;
	if (_.isEmpty(options)) {
		if (!argv.silent) console.log("Options are empty. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.updateMediaType(mediaTypeId, { text, presets, options });
}

function mediaTypeDelete(b, argv) {
	const mediaTypeId = argv.id;
	if (!mediaTypeId) {
		if (!argv.silent) console.log("No ID given for media type. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.deleteMediaType(mediaTypeId);
}

function actionList(b, argv) {
	return b.listActions();
}

function actionGet(b, argv) {
	const actionId = argv.id;
	if (!actionId) {
		if (!argv.silent) console.log("No ID given for action. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.getAction(actionId);
}

function actionAdd(b, argv) {
	const actionId = argv.id;
	if (!actionId) {
		if (!argv.silent) console.log("No ID given for action. Cancelling the request.");
		process.exit(1);
		return;
	}
	const action = argv.action;
	if (_.isEmpty(action)) {
		if (!argv.silent) console.log("Action is empty. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.addAction(actionId, action);
}

function actionUpdate(b, argv) {
	const actionId = argv.id;
	if (!actionId) {
		if (!argv.silent) console.log("No ID given for action. Cancelling the request.");
		process.exit(1);
		return;
	}
	const action = argv.action;
	if (_.isEmpty(action)) {
		if (!argv.silent) console.log("Action is empty. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.updateAction(actionId, action);
}

function actionDelete(b, argv) {
	const actionId = argv.id;
	if (!actionId) {
		if (!argv.silent) console.log("No ID given for action. Cancelling the request.");
		process.exit(1);
		return;
	}
	return b.deleteAction(actionId);
}

function exportAll(b, argv) {
	const theme = argv.theme;
	const zoom = argv.zoom;
	const subPath = argv.path;
	const logo = argv.logo;
	const header = argv.header;
	const tools = argv.tools;
	const toolbars = argv.toolbars;
	return b.exportPanels({
		path: subPath,
		theme, zoom, logo, header, tools, toolbars,
	});
}

function exportPanel(b, argv) {
	const panelId = argv.id;
	const theme = argv.theme;
	const zoom = argv.zoom;
	const subPath = argv.path;
	const filename = argv.name;
	const logo = argv.logo;
	const header = argv.header;
	const tools = argv.tools;
	const toolbars = argv.toolbars;
	return b.exportPanel(panelId, {
		path: subPath,
		name: filename,
		theme, zoom, logo, header, tools, toolbars,
	});
}

function exportSlot(b, argv) {
	const panelId = argv.panelId || 'default';
	const slotId = argv.slotId;
	if (!slotId) {
		if (!argv.silent) console.log("No slot ID given. Cancelling the request.");
		process.exit(1);
		return;
	}
	const theme = argv.theme;
	const zoom = argv.zoom;
	const subPath = argv.path;
	const filename = argv.name;
	const tools = argv.tools;
	const toolbar = argv.toolbar;
	return b.exportSlot(panelId, slotId, {
		path: subPath,
		name: filename,
		theme, zoom, tools, toolbar,
	});
}

function requestUrl(request) {
	return _.trimEnd(request.prefixUrl, '/') + '/' + _.trimStart(request.url, '/');
}

async function printRequestInfo(request) {
	console.log('The following request would have been sent.');
	console.log('');
	console.log('HTTP Method:', request.method);
	console.log('URL:        ', requestUrl(request));
	console.log('Headers:');
	_.forEach(request.headers, (v, k) => {
		if (k === 'Content-Type') return;
		console.log('    ', k + ':', v);
	});
	if (request.json) {
		console.log('    ', 'Content-Type:', 'application/json');
		console.log('Body:');
		console.log(JSON.stringify(request.json, null, '  '));
	} else if (request.body) {
		if (_.isString(request.body)) {
			console.log('    ', 'Content-Type:', 'text/plain');
			console.log('Body:');
			console.log(request.body);
		} else if (request.body === process.stdin) {
			console.log('    ', 'Content-Type:', request.headers['Content-Type']);
			console.log('Body: Data streamed from STDIN');
		} else if (isStream(request.body) && typeof request.body.path === 'string') {
			console.log('    ', 'Content-Type:', request.headers['Content-Type']);
			console.log('Body: Data streamed from', request.body.path);
		}
	}
}

function escapeDataForBourneShell(s) {
	// does only support Bourne like shells (bash etc.) for now
	// for Windows CMD and PowerShell another escaping is required each
	return (s
		.replace(/'/g, "'\\''")
	);
}

async function printCurlRequest(request) {
	let cmd = 'curl';
	if (request.method != 'GET') {
		cmd += ' -X ' + request.method;
	}
	_.forEach(request.headers, (v, k) => {
		cmd += ' -H "' + k + ': ' + v + '"';
	});
	if (request.json) {
		cmd += ' -H "Content-Type: application/json"';
		cmd += " -d '" + escapeDataForBourneShell(JSON.stringify(request.json)) + "'";
	} else if (request.body) {
		if (_.isString(request.body)) {
			cmd += ' -H "Content-Type: text/plain"';
			cmd += " -d '" + escapeDataForBourneShell(request.body) + "'";
		} else if (request.body === process.stdin) {
			cmd += ' --data-binary @-'
		} else if (isStream(request.body) && typeof request.body.path === 'string') {
			cmd += ' --data-binary "@' + request.body.path + '"'
		}
	}
	cmd += ' ' + requestUrl(request);
	console.log(cmd);
}

async function dryInterceptor(argv, options) {
	if (argv.curl)
		await printCurlRequest(options);
	else
		await printRequestInfo(options);
	return null;
}

const commands = {
	'display': display,
	'eval': evaluateCode,
	'panel ls': panelList,
	'panel get': panelGet,
	'panel add': panelAdd,
	'panel update': panelLayout,
	'panel layout': panelLayout,
	'panel clear': panelClear,
	'panel export': panelExport,
	'panel delete': panelDelete,
	'slot clear': slotClear,
	'slot export': slotExport,
	'preset ls': presetList,
	'preset get': presetGet,
	'preset add': presetAdd,
	'preset update': presetUpdate,
	'preset delete': presetDelete,
	'type ls': mediaTypeList,
	'type get': mediaTypeGet,
	'type add': mediaTypeAdd,
	'type update': mediaTypeUpdate,
	'type delete': mediaTypeDelete,
	'action ls': actionList,
	'action get': actionGet,
	'action add': actionAdd,
	'action update': actionUpdate,
	'action delete': actionDelete,
	'export all': exportAll,
	'export panel': exportPanel,
	'export slot': exportSlot,
};

async function run(argv, f) {
	const config = {
		server: { },
		client: {
			format: 'text/plain',
			token: argv.token,
		},
	};
	if (argv.api) config.server.url = argv.api;

	const b = await boomack.withConfig({
		loadDefaultFiles: true,
		additionalFiles: argv.configFiles,
		loadEnvironmentVars: true,
		config: config,
	});
	if (argv.dry) {
		b.setInterceptor(options => dryInterceptor(argv, options));
	}
	try {
		const response = await f(b, argv);
		if (response && !argv.dry) {
			console.log(response.body);
			return !!response.success;
		} else {
			return true;
		}
	} catch (err) {
		console.log(err);
		return false;
	}
}

exports.run = async function (argv) {
	const cmdFn = _.get(commands, argv.command);
	if (!_.isFunction(cmdFn)) throw new Error("Unknown command: " + argv.command);
	return await run(argv, cmdFn);
};
